# coding: utf-8

from contextlib import closing
from contextlib import contextmanager

from django.core.serializers import python as python_serializer_module
from django.core.serializers import deserialize
from django.db import connection
from django.db.migrations.operations.base import Operation
from django.db.models import Max


def get_objects_from_fixture(file_path, file_type=None):
    u"""Возвращает генератор объектов из файла фикстуры.

    :param basestring file_path: Путь к файлу с данными.
    :param basestring file_type: Тип файла фикстуры (xml, json или yaml).

    :rtype: generator
    """
    if file_type is None:
        file_type = file_path[file_path.rfind('.') + 1:]
        if file_type not in ('json', 'xml', 'yaml'):
            raise ValueError(u'Неподдерживаемый тип файла ' + file_path)

    with file(file_path, 'r') as infile:
        with closing(deserialize(file_type, infile.read())) as objects:
            for obj in objects:
                yield obj


def correct_sequence_value(model, field='id', conn=connection):
    u"""Корректирует значение последовательности для указанного поля модели.

    Устанавливает в качестве значения последовательности максимальное значение
    указанного поля. Актуально, когда, например, после загрузки какихм-либо
    данных из фикстур становится возможной ситуация, когда при добавлении
    очередной записи последовательность выдаёт значение, которое уже есть в БД.

    :param model: Класс модели.
    :param str field: Имя поля, для последовательности которого выполняется
        корректировка значения.
    :param conn: Подключение к БД, в которой размещена таблица указанной
        модели.
    """
    max_id = model.objects.aggregate(max_id=Max(field))['max_id'] or 1

    cursor = conn.cursor()

    cursor.execute(
        "SELECT setval(pg_get_serial_sequence(%s,%s), %s)", (
            model._meta.db_table,
            model._meta.get_field(field).column,
            max_id,
        )
    )


class LoadFixture(Operation):

    u"""Операция загрузки фикстур в миграции."""

    reversible = False

    reduces_to_sql = False

    atomic = True

    def __init__(self, file_path):
        self.file_path = file_path

    def state_forwards(self, app_label, state):
        pass

    def database_forwards(self, app_label, schema_editor, from_state,
                          to_state):
        @contextmanager
        def replace_model_loader():
            _get_model = python_serializer_module._get_model
            python_serializer_module._get_model = to_state.apps.get_model
            yield
            python_serializer_module._get_model = _get_model

        db_alias = schema_editor.connection.alias

        with replace_model_loader():
            for obj in get_objects_from_fixture(self.file_path):
                model = obj.object.__class__
                if self.allow_migrate_model(db_alias, model):
                    obj.save()


class CorrectSequence(Operation):

    u"""Корректирует значение последовательности для указанного поля.

    .. seealso::

        :func:`~educommon.django.db.migration.operations.correct_sequence_value`
    """

    reversible = False

    reduces_to_sql = False

    def __init__(self, model_name):
        self.model_name = model_name

    def state_forwards(self, app_label, state):
        pass

    def database_forwards(self, app_label, schema_editor, from_state,
                          to_state):
        db_alias = schema_editor.connection.alias
        model = to_state.apps.get_model(app_label, self.model_name)
        if self.allow_migrate_model(db_alias, model):
            correct_sequence_value(model, conn=schema_editor.connection)
